#include <R.h>
#include <Rinternals.h>
#include <Rdefines.h>
#include <Rmath.h>
#include <R_ext/Rdynload.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Auxiliary function to get the list element named str, or return NULL */

SEXP getListElementNu(SEXP list, const char *str) {

	SEXP elmt = R_NilValue, names = getAttrib(list, R_NamesSymbol);
	int i;

	for (i = 0; i < length(list); i++)
		if(strcmp(CHAR(STRING_ELT(names, i)), str) == 0) {
			elmt = VECTOR_ELT(list, i);
			break;
		}
	
	return elmt;
}

/**
 * Function which computes the t copula density.
 *
 */

double tDensityNu(double x, double y, double a, double b) {

	double rho, tau, nu;

	tau = 0.99 * (2 * pnorm(a, 0, 1, 1, 0) - 1);
	nu = 1 + 1e6 * pnorm(b, 0, 1, 1, 0);

	rho = sin((tau * PI) / 2);

	x = qt(x, nu, 1, 0);
	y = qt(y, nu, 1, 0);

	return 1 / (2 * PI * sqrt(1 - rho * rho)) *
		pow(1 + (x * x + y * y - 2 * rho * x * y) / (nu * (1 - rho * rho)), -(nu + 2) / 2) / dt(x, nu, 0) / dt(y, nu, 0);
}

/**
 * Function which computes the normalization constant.
 *
 */

void computeNormalizationConstantAndMomentsNuInternal(double X, double Y, double a, double mb, double vb,
	double *Z, double *m1, double *m2, int nSplits) {

	int i;
	double maxb, minb, hb, xb, sqrtvb;
	double densityValue, weight, factor;

	sqrtvb = sqrt(vb);

	maxb = mb + 6 * sqrtvb;
	minb = mb - 6 * sqrtvb;
	hb = (maxb - minb) / (nSplits * 4);
	
	*Z = *m1 = *m2 = 0;
	xb = minb;

	weight = 7 * hb * 2.0 / 45;
	densityValue = tDensityNu(X, Y, a, xb) * dnorm(xb, mb, sqrtvb, 0);
	*Z += densityValue * weight;
	*m1 += xb * densityValue * weight;
	*m2 += xb * xb * densityValue * weight;
	xb += hb;
	for (i = 0 ; i < nSplits ; i++) {

		weight = 32 * hb * 2.0 / 45;
		densityValue = tDensityNu(X, Y, a, xb) * dnorm(xb, mb, sqrtvb, 0);
		*Z += densityValue * weight;
		*m1 += xb * densityValue * weight;
		*m2 += xb * xb * densityValue * weight;
		xb += hb;

		weight = 12 * hb * 2.0 / 45;
		densityValue = tDensityNu(X, Y, a, xb) * dnorm(xb, mb, sqrtvb, 0);
		*Z += densityValue * weight;
		*m1 += xb * densityValue * weight;
		*m2 += xb * xb * densityValue * weight;
		xb += hb;

		weight = 32 * hb * 2.0 / 45;
		densityValue = tDensityNu(X, Y, a, xb) * dnorm(xb, mb, sqrtvb, 0);
		*Z += densityValue * weight;
		*m1 += xb * densityValue * weight;
		*m2 += xb * xb * densityValue * weight;
		xb += hb;

		if (i != nSplits - 1)
			factor = 2;
		else
			factor = 1;

		weight = 7 * hb * 2.0 / 45;
		densityValue = tDensityNu(X, Y, a, xb) * dnorm(xb, mb, sqrtvb, 0);
		*Z += factor * densityValue * weight;
		*m1 += factor * xb * densityValue * weight;
		*m2 += factor * xb * xb * densityValue * weight;
		xb += hb;
	}
}

/**
 * Function which computes the normalization constant.
 *
 */

void computeNormalizationConstantAndMomentsNuExternal(double X, double Y, double ma, double va, double mb, double vb,
	double *Z, double *m1, double *m2, int nSplits) {

	/* We get the grid points */

	double sqrtva;
	double x1, x2, x3, x4, x5;
	double w1, w2, w3, w4, w5;
	double z1, z2, z3, z4, z5;
	double m11, m12, m13, m14, m15;
	double m21, m22, m23, m24, m25;

	sqrtva = sqrt(va);
	
	x1 = ma - 3 * sqrtva;
	x2 = ma - 1.5 * sqrtva;
	x3 = ma;
	x2 = ma + 1.5 * sqrtva;
	x5 = ma + 3 * sqrtva;

	w1 = (x5 - x1) / 90 * 7;
	w2 = (x5 - x1) / 90 * 32;
	w3 = (x5 - x1) / 90 * 12;
	w4 = (x5 - x1) / 90 * 32;
	w5 = (x5 - x1) / 90 * 7;

	computeNormalizationConstantAndMomentsNuInternal(X, Y, x1, mb, vb, &z1, &m11, &m21, nSplits);
	computeNormalizationConstantAndMomentsNuInternal(X, Y, x2, mb, vb, &z2, &m12, &m22, nSplits);
	computeNormalizationConstantAndMomentsNuInternal(X, Y, x3, mb, vb, &z3, &m13, &m23, nSplits);
	computeNormalizationConstantAndMomentsNuInternal(X, Y, x4, mb, vb, &z4, &m14, &m24, nSplits);
	computeNormalizationConstantAndMomentsNuInternal(X, Y, x5, mb, vb, &z5, &m15, &m25, nSplits);

	z1 *= w1 * dnorm(x1, ma, sqrtva, 0);
	z2 *= w2 * dnorm(x2, ma, sqrtva, 0);
	z3 *= w3 * dnorm(x3, ma, sqrtva, 0);
	z4 *= w4 * dnorm(x4, ma, sqrtva, 0);
	z5 *= w5 * dnorm(x5, ma, sqrtva, 0);

	m11 *= w1 * dnorm(x1, ma, sqrtva, 0);
	m12 *= w2 * dnorm(x2, ma, sqrtva, 0);
	m13 *= w3 * dnorm(x3, ma, sqrtva, 0);
	m14 *= w4 * dnorm(x4, ma, sqrtva, 0);
	m15 *= w5 * dnorm(x5, ma, sqrtva, 0);

	m21 *= w1 * dnorm(x1, ma, sqrtva, 0);
	m22 *= w2 * dnorm(x2, ma, sqrtva, 0);
	m23 *= w3 * dnorm(x3, ma, sqrtva, 0);
	m24 *= w4 * dnorm(x4, ma, sqrtva, 0);
	m25 *= w5 * dnorm(x5, ma, sqrtva, 0);

	*Z = z1 + z2 + z3 + z4 + z5;
	*m1 = m11 + m12 + m13 + m14 + m15;
	*m2 = m21 + m22 + m23 + m24 + m25;
}

/**
 * Function which refines the likelihood factors.
 *
 */

SEXP refineLikelihoodFastNu(SEXP R_m) {

	/* Auxiliary variables */

	int i;
	double Z, m1, m2;
	double maNew, vaNew, eta1New, eta2New, eta1HatNew, eta2HatNew;

	/* Local points to the variables stored in R_m */

	double *X;
	double *Y;
	double *ma;
	double *va;
	double *f1HatEta1;
	double *f1HatEta2;
	double *mb;
	double *vb;
	double *logZ;
	double damping;
	int n;

	/* We access the elements from the list */

	X = NUMERIC_POINTER(getListElementNu(R_m, "X"));
	Y = NUMERIC_POINTER(getListElementNu(R_m, "Y"));
	ma = NUMERIC_POINTER(getListElementNu(R_m, "ma"));
	va = NUMERIC_POINTER(getListElementNu(R_m, "va"));
	f1HatEta1 = NUMERIC_POINTER(getListElementNu(R_m, "f1HatEta1"));
	f1HatEta2 = NUMERIC_POINTER(getListElementNu(R_m, "f1HatEta2"));
	mb = NUMERIC_POINTER(getListElementNu(R_m, "mb"));
	vb = NUMERIC_POINTER(getListElementNu(R_m, "vb"));
	logZ = NUMERIC_POINTER(getListElementNu(R_m, "logZ"));
	damping = *NUMERIC_POINTER(getListElementNu(R_m, "damping"));
	n = *INTEGER_POINTER(getListElementNu(R_m, "n"));

	/* Main loop */

	for (i = 0 ; i < n ; i++) {

		if (va[ i ] > 0 && vb[ i ]  > 0) {

			/* We compute the normalization contant */

//			computeNormalizationConstantAndMomentsNuExternal(X[ i ], Y[ i ], ma[ i ], va[ i ], mb[ i ], vb[ i ], &Z, &m1, &m2, 8);
			computeNormalizationConstantAndMomentsNuInternal(X[ i ], Y[ i ], ma[ i ], mb[ i ], vb[ i ], &Z, &m1, &m2, 10);

			if (Z > 1e-10) {

				maNew = m1 / Z;
				vaNew = m2 / Z - maNew * maNew;

				eta1New = maNew / vaNew;
				eta2New = 1 / vaNew;
				eta1HatNew = eta1New - mb[ i ] / vb[ i ];
				eta2HatNew = eta2New - 1 / vb[ i ];

				if (eta2HatNew < 1e10 && eta1HatNew < 1e10 && eta1HatNew > -1e10) {
					f1HatEta2[ i ] = damping * eta2HatNew + (1 - damping) * f1HatEta2[ i ];
					f1HatEta1[ i ] = damping * eta1HatNew + (1 - damping) * f1HatEta1[ i ];
					logZ[ i ] = log(Z);
				}
			}	
		}
	}

	return R_m;
}
