#include <R.h>
#include <Rinternals.h>
#include <Rdefines.h>
#include <Rmath.h>
#include <R_ext/Rdynload.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Auxiliary function to get the list element named str, or return NULL */

SEXP getListElementTau(SEXP list, const char *str) {

	SEXP elmt = R_NilValue, names = getAttrib(list, R_NamesSymbol);
	int i;

	for (i = 0; i < length(list); i++)
		if(strcmp(CHAR(STRING_ELT(names, i)), str) == 0) {
			elmt = VECTOR_ELT(list, i);
			break;
		}
	
	return elmt;
}

/**
 * Function which computes the t copula density.
 *
 */

double tDensityTau(double x, double y, double a, double b) {

	double rho, tau, nu;

	tau = 0.99 * (2 * pnorm(a, 0, 1, 1, 0) - 1);
	nu = 1 + 1e6 * pnorm(b, 0, 1, 1, 0);

	rho = sin((tau * PI) / 2);

	x = qt(x, nu, 1, 0);
	y = qt(y, nu, 1, 0);

	return 1 / (2 * PI * sqrt(1 - rho * rho)) *
		pow(1 + (x * x + y * y - 2 * rho * x * y) / (nu * (1 - rho * rho)), -(nu + 2) / 2) / dt(x, nu, 0) / dt(y, nu, 0);
}

/**
 * Function which computes the normalization constant.
 *
 */

void computeNormalizationConstantAndMomentsTau(double X, double Y, double a, double mb, double vb,
	double *Z, double *m1, double *m2, int nSplits) {

	int i;
	double maxb, minb, hb, xb, sqrtvb;
	double densityValue, weight, factor;

	sqrtvb = sqrt(vb);

	maxb = mb + 6 * sqrtvb;
	minb = mb - 6 * sqrtvb;
	hb = (maxb - minb) / (nSplits * 4);
	
	*Z = *m1 = *m2 = 0;
	xb = minb;

	weight = 7 * hb * 2.0 / 45;
	densityValue = tDensityTau(X, Y, xb, a) * dnorm(xb, mb, sqrtvb, 0);
	*Z += densityValue * weight;
	*m1 += xb * densityValue * weight;
	*m2 += xb * xb * densityValue * weight;
	xb += hb;
	for (i = 0 ; i < nSplits ; i++) {

		weight = 32 * hb * 2.0 / 45;
		densityValue = tDensityTau(X, Y, xb, a) * dnorm(xb, mb, sqrtvb, 0);
		*Z += densityValue * weight;
		*m1 += xb * densityValue * weight;
		*m2 += xb * xb * densityValue * weight;
		xb += hb;

		weight = 12 * hb * 2.0 / 45;
		densityValue = tDensityTau(X, Y, xb, a) * dnorm(xb, mb, sqrtvb, 0);
		*Z += densityValue * weight;
		*m1 += xb * densityValue * weight;
		*m2 += xb * xb * densityValue * weight;
		xb += hb;

		weight = 32 * hb * 2.0 / 45;
		densityValue = tDensityTau(X, Y, xb, a) * dnorm(xb, mb, sqrtvb, 0);
		*Z += densityValue * weight;
		*m1 += xb * densityValue * weight;
		*m2 += xb * xb * densityValue * weight;
		xb += hb;

		if (i != nSplits - 1)
			factor = 2;
		else
			factor = 1;

		weight = 7 * hb * 2.0 / 45;
		densityValue = tDensityTau(X, Y, xb, a) * dnorm(xb, mb, sqrtvb, 0);
		*Z += factor * densityValue * weight;
		*m1 += factor * xb * densityValue * weight;
		*m2 += factor * xb * xb * densityValue * weight;
		xb += hb;
	}
}

/**
 * Function which refines the likelihood factors.
 *
 */

SEXP refineLikelihoodFastTau(SEXP R_m) {

	/* Auxiliary variables */

	int i;
	double Z, m1, m2;
	double maNew, vaNew, eta1New, eta2New, eta1HatNew, eta2HatNew;

	/* Local points to the variables stored in R_m */

	double *X;
	double *Y;
	double *ma;
	double *va;
	double *f1HatEta1;
	double *f1HatEta2;
	double *mb;
	double *vb;
	double *logZ;
	double damping;
	int n;

	/* We access the elements from the list */

	X = NUMERIC_POINTER(getListElementTau(R_m, "X"));
	Y = NUMERIC_POINTER(getListElementTau(R_m, "Y"));
	ma = NUMERIC_POINTER(getListElementTau(R_m, "ma"));
	va = NUMERIC_POINTER(getListElementTau(R_m, "va"));
	f1HatEta1 = NUMERIC_POINTER(getListElementTau(R_m, "f1HatEta1"));
	f1HatEta2 = NUMERIC_POINTER(getListElementTau(R_m, "f1HatEta2"));
	mb = NUMERIC_POINTER(getListElementTau(R_m, "mb"));
	vb = NUMERIC_POINTER(getListElementTau(R_m, "vb"));
	logZ = NUMERIC_POINTER(getListElementTau(R_m, "logZ"));
	damping = *NUMERIC_POINTER(getListElementTau(R_m, "damping"));
	n = *INTEGER_POINTER(getListElementTau(R_m, "n"));

	/* Main loop */

	for (i = 0 ; i < n ; i++) {

		if (va[ i ] > 0 && vb[ i ]  > 0) {

			/* We compute the normalization contant */

			computeNormalizationConstantAndMomentsTau(X[ i ], Y[ i ], mb[ i ], ma[ i ], va[ i ], &Z, &m1, &m2, 10);

			if (Z > 1e-10) {

				maNew = m1 / Z;
				vaNew = m2 / Z - maNew * maNew;

				eta1New = maNew / vaNew;
				eta2New = 1 / vaNew;
				eta1HatNew = eta1New - ma[ i ] / va[ i ];
				eta2HatNew = eta2New - 1 / va[ i ];

				if (eta2HatNew < 1e10 && eta1HatNew < 1e10 && eta1HatNew > -1e10) {
					f1HatEta2[ i ] = damping * eta2HatNew + (1 - damping) * f1HatEta2[ i ];
					f1HatEta1[ i ] = damping * eta1HatNew + (1 - damping) * f1HatEta1[ i ];
					logZ[ i ] = log(Z);
				}
			}	
		}
	}

	return R_m;
}
